/*
 *         bpack - A Burrows-Wheeler Transform and Huffman Encoding
 *                            Implementation
 *
 *  Copyright (C) 2006      Vasco Costa <the_papon@hotmail.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "mtf.h"

void mtf_encode(unsigned char *buf_in, unsigned char *buf_out, int tam){
	int i,j;
	unsigned char list[256];
     
	for(i=0; i<256; i++) list[i]=i;
         
	for(i=0; i<tam; i++){
		buf_out[i] = list[buf_in[i]];
        
		for(j=0; j<256; j++)
			if(list[j] < list[buf_in[i]])
				list[j]++;	
		list[buf_in[i]]=0;
	}
}        
        
void mtf_decode(unsigned char *buf_in, unsigned char *buf_out, int tam){
     int i,j;
     unsigned char list[256];
     unsigned char aux;
     
	for(i=0; i<256; i++) list[i]=i;
         
	for(i=0; i<tam; i++){
		buf_out[i]= list[buf_in[i]];
		aux=list[buf_in[i]];

		for(j=buf_in[i]; j!=0; j--) list[j]=list[j-1];
		list[0]=aux;
	}
}
